package nDBI;

$VERSION = '0.0.8';
use nDBI::Statement;
my $int_error = '';
1;

sub connect {
	my $obj = shift;	
	my $dr = $_[0];
	
	require DBI;
	@ISA = (DBI);
	
	$dr =~ s/^dbi://;  
	$dr =~ s/^.*dsn=DBI:// if $dr =~ /^Proxy/; 		
	$dr =~ s/:.*//;  my $drr = "nDBD/$dr.pm";
	require $drr;  $dr = "nDBD::$dr";
	my $dbi;
	
	if (defined ($dbi = eval { DBI->connect(@_) }) ) {
		my $ndbd = new $dr $dbi;
		return bless { dbi => $dbi, ndbd => $ndbd },$obj;
	} else {
		return undef;
	}
}

sub proxy {
	my $obj = shift;  my $host = shift;  my $port = shift;
	
	require IO::Socket;
	require Storable;  
	
	$int_error = '';
	my $sock = new IO::Socket::INET(PeerHost => $host || 'localhost',
	                 PeerPort => $port || 2401, Proto => 'tcp');
	if (defined $sock) {
		$obj = bless { proxy => 1, sock => $sock },$obj;	
		$res = $obj->proxy_get('connect',0,@_);
		if (defined $res) {
			$int_error = $res;
			return undef;
		}
		return $obj;
	} else {
		$int_error = $!;
		return undef;
	}
}

sub _rcv {
	my $sock = shift;  my $buf = '';  my $length = 0;  my $offset = 0;
	my $bytes_read = sysread($sock,$buf,4);
	unless ($! || $bytes_read != 4) { $length = unpack('N',$buf); }

	while ($length > 0) {
		$bytes_read = sysread($sock,$buf,$length,$offset);
		if (defined $bytes_read) {
			last unless $bytes_read;	
			$length -= $bytes_read;  $offset += $bytes_read;
		} else { last; }
	}
	return undef if $length > 0;
	eval {
		$buf = Storable::thaw($buf);
	};  if ($@) { $bug = ''; }
	return $buf;
}

sub _send {
	my ($sock,$msg) = @_;  $msg = Storable::nfreeze($msg);
      	my $offset = 0;  my $length = length $msg;  
	$msg = pack('N',$length) . $msg;  $length = length $msg;

	while ($length > 0) {
		my $bytes_written = syswrite($sock,$msg,$length,$offset);
		$sock->flush();
		if (defined $bytes_written) {
			last unless $bytes_written;	
			$length -= $bytes_written;  $offset += $bytes_written;
		} else { last; }
	}
	return not $length;
}

sub proxy_st_get {
	my ($obj,$id,$func,$wantarray,@params) = @_;	
	delete $obj->{statements}->{$id} if $func eq 'finish';
	$func = 'nDBI::st::' . $id . "::$func";
	return $obj->proxy_get($func,$wantarray,@params);
}

sub proxy_get {
	my ($obj,$func,$wantarray,@params) = @_;
	_send $obj->{sock}, [ $func,$wantarray,@params ];
	my $r = _rcv $obj->{sock};
	if ($wantarray) { return @$$r; } else { 
		$r = $$r;
		if ($r =~ 'DBI::st::(.*)') {
			unless (exists $obj->{statements}->{$1}) {
				$obj->{statements}->{$1} = 
					new nDBI::Statement $1,$obj;
			}
			$r = $obj->{statements}->{$1};
		}
		return $r; 
	}
}

sub AUTOLOAD {
	my $obj = shift;  my @ARGS = @_;
	$AUTOLOAD =~ s/.*:://;
	if ($obj->{proxy}) {
		if ($AUTOLOAD eq 'disconnect') {
			my $r = $obj->proxy_get($AUTOLOAD,0,@ARGS);
			close $obj->{sock};
			return $r;
		}
		return $obj->proxy_get($AUTOLOAD,wantarray,@ARGS);
	} else {
		if ($AUTOLOAD =~ /^ndbd_/) {
			return $obj->{ndbd}->$AUTOLOAD(@ARGS);
		} else {
			return $obj->{dbi}->$AUTOLOAD(@ARGS);
		}
	}
}

sub all_drivers {
	my %seen_dir = ();  my %seen_ndbd = ();
	for (@INC) {
		next unless -d "$_/nDBD";
		next if $seen_dir{$_};
		++$seen_dir{$_};
		if (opendir(DIR,"$_/nDBD")) {
			for (sort readdir DIR) {
				next unless s/\.pm$//;
				++$seen_ndbd{$_};
			}
			closedir DIR;
		}
	}
	return sort keys %seen_ndbd;
}

sub available_drivers {
	my @drivers = ();
	my %seen_dir;  my %seen_ndbd;
	for (@INC) {
		next unless -d "$_/nDBD";
		next if $seen_dir{$_};
		++$seen_dir{$_};
		if (opendir(DIR,"$_/nDBD")) {
			for (sort readdir DIR) {
				next unless s/\.pm$//;
				++$seen_ndbd{$_};
			}
			closedir DIR;
		}
	}
	require DBI;
	for (DBI->available_drivers()) {
		push @drivers,$_ if $seen_ndbd{$_}; 
	}
	return @drivers;
}

sub errstr {
	my $obj = shift;
	if ($int_error) { my $e = $int_error;  $int_error = '';  return $e; }
	return $obj->proxy_get('errstr',0) if $obj->{proxy};
	return $DBI::errstr;	
}

sub looks_like_number {
	my $what = shift;
	return $what =~ /^(?:\d+(?:\.\d+)?|\.\d+)$/;
}

sub ndbi_set {
	my $obj = shift;
	return $obj->proxy_get('ndbi_set',0,@_) if $obj->{proxy};
	my %what = @_;
	for (keys %what) { $obj->{dbi}->{$_} = $what{$_}; }
}

sub ndbi_login_string {
	my ($dbsys,$database,$host,$port) = @_;
	eval {
		my $drr = "nDBD/$dbsys.pm";  require $drr;  
	};
	return $database if $@;
	
	my $fce = 'nDBD::'.$dbsys.'::ndbd_login_string';
	my $res = &$fce ($database,$host,$port);	
	return $res?$res:$database;
}

sub ndbi_tables {
	my $obj = shift;
	my $ref = $obj->ndbd_tables();
	if (defined $ref) {
		return @$ref;
	} else {
		my @tab = $obj->tables;
		if (not @tab) {
			return $obj->func('_ListTables');
		} else {
			return @tab;
		}
	}
}
